"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Transporter = void 0;
const sink_1 = require("../../sink");
const bayLog_1 = require("../../bayLog");
const ioException_1 = require("../../util/ioException");
const upgradeException_1 = require("../upgradeException");
const protocolException_1 = require("../../protocol/protocolException");
const nextSocketAction_1 = require("../nextSocketAction");
class Transporter {
    constructor(serverMode, traceSSL, writeOnly = false) {
        this.dataListener = null;
        this.serverMode = false;
        this.traceSSL = false;
        this.channel = null; // socket or file
        this.finale = false;
        this.initialized = false;
        this.chValid = false;
        this.needHandshake = true;
        this.nonBlockingHandler = null;
        this.tmpAddress = [];
        this.readBuf = null;
        this.serverMode = serverMode;
        this.traceSSL = traceSSL;
        this.readBuf = Buffer.alloc(8192);
        this.writeOnly = writeOnly;
    }
    init(chHnd, ch, lis) {
        if (this.initialized)
            throw new sink_1.Sink(this + " This transporter is already in use by channel: " + ch);
        this.nonBlockingHandler = chHnd;
        this.dataListener = lis;
        this.channel = ch;
        this.setValid(true);
        this.initialized = true;
        this.nonBlockingHandler.addChannelListener(ch, this);
    }
    toString() {
        return "tp[" + this.dataListener + "]";
    }
    //////////////////////////////////////////////////////
    // implements Reusable
    //////////////////////////////////////////////////////
    reset() {
        this.finale = false;
        this.initialized = false;
        this.channel = null;
        this.setValid(false);
        this.needHandshake = true;
    }
    //////////////////////////////////////////////////////
    // implements Postman
    //////////////////////////////////////////////////////
    post(buf, adr, tag, listener) {
        this.checkInitialized();
        bayLog_1.BayLog.debug("%s post: %s len=%d", this, tag, buf.length);
        if (!this.chValid) {
            throw new ioException_1.IOException("Invalid channel");
        }
        else {
            try {
                this.nonBlockingHandler.askToWrite(buf, this.channel, listener);
            }
            catch (e) {
                this.chValid = false;
                throw e;
            }
        }
    }
    //////////////////////////////////////////////////////
    // implements Valve
    //////////////////////////////////////////////////////
    openValve() {
        bayLog_1.BayLog.debug("%s resume", this);
        this.nonBlockingHandler.askToRead(this.channel);
    }
    //////////////////////////////////////////////////////
    // implements ChannelListener
    //////////////////////////////////////////////////////
    onConnect(chkCh) {
        this.checkChannel(chkCh);
        bayLog_1.BayLog.trace("%s onConnect", this);
        return this.dataListener.notifyConnect();
    }
    onRead(ch, buf) {
        this.checkChannel(ch);
        bayLog_1.BayLog.trace("%s onRead", this);
        try {
            if (buf == null || buf.length == 0) {
                return this.dataListener.notifyEof();
            }
            else {
                try {
                    return this.dataListener.notifyRead(buf, null);
                }
                catch (e) {
                    if (e instanceof upgradeException_1.UpgradeException) {
                        bayLog_1.BayLog.debug("%s Protocol upgrade", this.dataListener);
                        return this.dataListener.notifyRead(buf, null);
                    }
                    else {
                        throw e;
                    }
                }
            }
        }
        catch (e) {
            if (e instanceof protocolException_1.ProtocolException) {
                let close = this.dataListener.notifyProtocolError(e);
                if (!close && this.serverMode)
                    return nextSocketAction_1.NextSocketAction.CONTINUE;
                else
                    return nextSocketAction_1.NextSocketAction.CLOSE;
            }
            else {
                this.onError(ch, e);
                return nextSocketAction_1.NextSocketAction.CLOSE;
            }
        }
    }
    onWrite(ch, listener) {
        this.checkChannel(ch);
        bayLog_1.BayLog.trace("%s onWrite", this);
        if (listener != null)
            listener();
        if (this.finale) {
            bayLog_1.BayLog.trace("%s finale and close", this);
            return nextSocketAction_1.NextSocketAction.CLOSE;
        }
        else if (this.writeOnly) {
            return nextSocketAction_1.NextSocketAction.SUSPEND;
        }
        else {
            return nextSocketAction_1.NextSocketAction.CONTINUE;
        }
    }
    checkTimeout(chkCh, durationSec) {
        this.checkChannel(chkCh);
        return this.dataListener.checkTimeout(durationSec);
    }
    onError(chkCh, e) {
        this.checkChannel(chkCh);
        //BayLog.error_e(e, "%s Error", this)
        this.dataListener.notifyError(e);
        this.chValid = false;
    }
    onClosed(chkCh) {
        try {
            this.checkChannel(chkCh);
        }
        catch (e) {
            bayLog_1.BayLog.error_e(e);
            return;
        }
        if (!this.chValid)
            return;
        this.setValid(false);
        this.dataListener.notifyClose();
    }
    getBuf() {
        return this.readBuf;
    }
    //////////////////////////////////////////////////////
    // other methods
    //////////////////////////////////////////////////////
    abort() {
        bayLog_1.BayLog.debug("%s abort", this);
        this.nonBlockingHandler.askToClose(this.channel);
    }
    isZombie() {
        return this.channel != null && !this.chValid;
    }
    flush() {
        this.checkInitialized();
        bayLog_1.BayLog.debug("%s flush", this);
    }
    postEnd() {
        this.checkInitialized();
        bayLog_1.BayLog.debug("%s postEnd vld=%s", this, this.chValid);
        this.finale = true;
    }
    checkChannel(chkCh) {
        this.checkInitialized();
        if (chkCh !== this.channel)
            throw new sink_1.Sink("%s Invalid transporter instance (ship was returned?)", this);
    }
    checkInitialized() {
        if (!this.initialized)
            throw new sink_1.Sink("%s Transporter not initialized", this);
    }
    setValid(valid) {
        this.chValid = valid;
    }
}
exports.Transporter = Transporter;
//# sourceMappingURL=transporter.js.map