"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BcfParser = void 0;
const bayMessage_1 = require("../bayMessage");
const symbol_1 = require("../symbol");
const arrayUtil_1 = require("../util/arrayUtil");
const bcfDocument_1 = require("./bcfDocument");
const parse_exception_1 = require("./parse_exception");
const fs_1 = require("fs");
const bcfElement_1 = require("./bcfElement");
const bcfKeyVal_1 = require("./bcfKeyVal");
class LineInfo {
    constructor(lineObj, indent) {
        this.lineObj = lineObj;
        this.indent = indent;
    }
}
class BcfParser {
    constructor() {
        this.indentMap = [];
        this.lineNo = 0;
    }
    pushIndent(spCount) {
        this.indentMap.push(spCount);
    }
    popIndent() {
        this.indentMap.splice(this.indentMap.length - 1, 1);
    }
    getIndent(spCount) {
        if (arrayUtil_1.ArrayUtil.empty(this.indentMap))
            this.pushIndent(spCount);
        else if (spCount > this.indentMap[this.indentMap.length - 1]) {
            this.pushIndent(spCount);
        }
        let indent = this.indentMap.indexOf(spCount);
        if (indent == -1)
            throw new parse_exception_1.ParseException(this.fileName, this.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.PAS_INVALID_INDENT));
        return indent;
    }
    parse(file) {
        let doc = new bcfDocument_1.BcfDocument();
        let currentContentList = doc.contentList;
        let parentContentList = null;
        this.fileName = file;
        this.lineNo = 0;
        try {
            this.input = (0, fs_1.readFileSync)(file).toString().split("\n");
            this.parseSameLevel(doc.contentList, 0);
        }
        catch (e) {
            if (e instanceof parse_exception_1.ParseException)
                throw e;
            else
                throw new parse_exception_1.ParseException(file, this.lineNo, e);
        }
        finally {
        }
        return doc;
    }
    parseSameLevel(curList, indent) {
        let objectExistsInSameLevel = false;
        while (true) {
            var lineInfo;
            if (this.prevLineInfo != null) {
                lineInfo = this.prevLineInfo;
                this.prevLineInfo = null;
                //BayServer.debug("returned line=" + lineInfo.lineObj + " indent=" + lineInfo.indent + " cur=" + indent);
            }
            else {
                if (this.lineNo == this.input.length)
                    break;
                let line = this.input[this.lineNo++];
                if (line.trim().startsWith("#") || line.trim() == "") {
                    // comment or empty line
                    continue;
                }
                lineInfo = this.parseLine(this.lineNo, line);
                //BayServer.debug("line=" + line + " indent=" + lineInfo.indent + " cur=" + indent);
            }
            if (lineInfo == null) {
                // comment or empty
                continue;
            }
            else if (lineInfo.indent > indent) {
                // lower level
                throw new parse_exception_1.ParseException(this.fileName, this.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.PAS_INVALID_INDENT));
            }
            else if (lineInfo.indent < indent) {
                // upper level
                this.prevLineInfo = lineInfo;
                if (objectExistsInSameLevel)
                    this.popIndent();
                return lineInfo;
            }
            else {
                objectExistsInSameLevel = true;
                // same level
                if (lineInfo.lineObj instanceof bcfElement_1.BcfElement) {
                    curList.push(lineInfo.lineObj);
                    let elm = lineInfo.lineObj;
                    let lastLineInfo = this.parseSameLevel(elm.contentList, lineInfo.indent + 1);
                    if (lastLineInfo == null) {
                        // EOF
                        this.popIndent();
                        return null;
                    }
                    else {
                        // Same level
                        continue;
                    }
                }
                else {
                    curList.push(lineInfo.lineObj);
                    continue;
                }
            }
        }
        this.popIndent();
        return null;
    }
    parseLine(lineNo, line) {
        var spCount;
        for (spCount = 0; spCount < line.length; spCount++) {
            let c = line[spCount];
            if (c.trim() != "")
                break;
            if (c != ' ')
                throw new parse_exception_1.ParseException(this.fileName, lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.PAS_INVALID_WHITESPACE));
        }
        let indent = this.getIndent(spCount);
        line = line.substring(spCount);
        line = line.trim();
        if (line.startsWith("[")) {
            let closePos = line.indexOf("]");
            if (closePos == -1) {
                throw new parse_exception_1.ParseException(this.fileName, lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.PAS_BRACE_NOT_CLOSED));
            }
            if (!line.endsWith("]")) {
                throw new parse_exception_1.ParseException(this.fileName, lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.PAS_INVALID_LINE));
            }
            let keyVal = this.parseKeyVal(line.substring(1, closePos), lineNo);
            return new LineInfo(new bcfElement_1.BcfElement(keyVal.key, keyVal.value, this.fileName, lineNo), indent);
        }
        else {
            return new LineInfo(this.parseKeyVal(line, lineNo), indent);
        }
    }
    parseKeyVal(line, lineNo) {
        let spPos = line.indexOf(' ');
        let key = spPos == -1 ? line : line.substring(0, spPos);
        let val = spPos == -1 ? "" : line.substring(spPos).trim();
        return new bcfKeyVal_1.BcfKeyVal(key, val, this.fileName, lineNo);
    }
}
exports.BcfParser = BcfParser;
//# sourceMappingURL=bcfParser.js.map