"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.InboundShip = void 0;
const ship_1 = require("../../watercraft/ship");
const counter_1 = require("../../util/counter");
const tourStore_1 = require("../../tour/tourStore");
const tour_1 = require("../../tour/tour");
const bayserver_1 = require("../../bayserver");
const sink_1 = require("../../sink");
const bayLog_1 = require("../../bayLog");
const arrayUtil_1 = require("../../util/arrayUtil");
const trouble_1 = require("../trouble");
const ioException_1 = require("../../util/ioException");
const httpException_1 = require("../../httpException");
const httpHeaders_1 = require("../../util/httpHeaders");
const buffer_1 = require("buffer");
const strUtil_1 = require("../../util/strUtil");
const httpStatus_1 = require("../../util/httpStatus");
class InboundShip extends ship_1.Ship {
    constructor() {
        super(...arguments);
        this.activeTours = [];
    }
    reset() {
        super.reset();
        if (!arrayUtil_1.ArrayUtil.empty(this.activeTours)) {
            throw new sink_1.Sink("%s There are some running tours", this);
        }
        this.needEnd = false;
    }
    initInbound(ch, agt, pm, portDkr, protoHandler) {
        super.init(ch, agt, pm);
        this.portDocker = portDkr;
        this.socketTimeoutSec = portDkr.getTimeoutSec() >= 0 ? portDkr.getTimeoutSec() : bayserver_1.BayServer.harbor.getSocketTimeoutSec();
        this.tourStore = tourStore_1.TourStore.getStore(agt.agentId);
        this.setProtocolHandler(protoHandler);
    }
    toString() {
        return this.agent + " ship#" + this.shipId + "/" + this.objectId + "[" + this.protocol() + "]";
    }
    //////////////////////////////////////////////////////
    // implements Reusable
    //////////////////////////////////////////////////////
    endShip() {
        bayLog_1.BayLog.debug("%s endShip", this);
        this.portDocker.returnProtocolHandler(this.agent, this.protocolHandler);
        this.portDocker.returnShip(this);
    }
    abortTours() {
        let returnList = [];
        // Abort tours
        for (const tur of this.activeTours) {
            if (tur.isValid()) {
                bayLog_1.BayLog.debug("%s is valid, abort it: stat=%s", tur, tur.state);
                if (tur.req.abort()) {
                    bayLog_1.BayLog.debug("%s aborted", tur);
                    returnList.push(tur);
                }
            }
        }
        for (const tur of returnList) {
            this.returnTour(tur);
        }
    }
    getTour(turKey, force = false) {
        if (turKey < 0 || turKey >= 0x1000000)
            throw new Error("Invalid tour key: " + turKey);
        let storeKey = InboundShip.uniqKey(this.shipId, turKey);
        var tur = this.tourStore.get(storeKey);
        if (tur == null) {
            tur = this.tourStore.rent(storeKey, force);
            if (tur == null)
                return null;
            tur.init(turKey, this);
            this.activeTours.push(tur);
        }
        if (tur.ship != this)
            throw new sink_1.Sink();
        tur.checkTourId(tur.id());
        return tur;
    }
    static uniqKey(sipId, turKey) {
        return sipId << 24 | (turKey & 0xffffffff);
    }
    returnTour(tur) {
        bayLog_1.BayLog.debug("%s Return tour: %s", this, tur);
        if (!(this.activeTours.includes(tur)))
            throw new sink_1.Sink("Tour is not in active list: %s", tur);
        this.tourStore.Return(InboundShip.uniqKey(this.shipId, tur.req.key));
        arrayUtil_1.ArrayUtil.remove(tur, this.activeTours);
        if (this.needEnd && arrayUtil_1.ArrayUtil.empty(this.activeTours)) {
            this.endShip();
        }
    }
    getErrorTour() {
        let turKey = InboundShip.errCounter.next();
        let storeKey = InboundShip.uniqKey(this.shipId, -turKey);
        let tur = this.tourStore.rent(storeKey, true);
        tur.init(-turKey, this);
        this.activeTours.push(tur);
        return tur;
    }
    sendHeaders(chkId, tur) {
        this.checkShipId(chkId);
        if (tur.isZombie() || tur.isAborted()) {
            // Don't send peer any data
            return;
        }
        var handled = false;
        if (!tur.errorHandling && tur.res.headers.status >= 400) {
            let trb = bayserver_1.BayServer.harbor.getTrouble();
            if (trb != null) {
                let cmd = trb.find(tur.res.headers.status);
                if (cmd != null) {
                    let errTour = this.getErrorTour();
                    errTour.req.uri = cmd.target;
                    tur.req.headers.copyTo(errTour.req.headers);
                    tur.res.headers.copyTo(errTour.res.headers);
                    errTour.req.remotePort = tur.req.remotePort;
                    errTour.req.remoteAddress = tur.req.remoteAddress;
                    errTour.req.serverAddress = tur.req.serverAddress;
                    errTour.req.serverPort = tur.req.serverPort;
                    errTour.req.serverName = tur.req.serverName;
                    errTour.res.headerSent = tur.res.headerSent;
                    tur.changeState(tour_1.Tour.TOUR_ID_NOCHECK, tour_1.Tour.STATE_ZOMBIE);
                    switch (cmd.method) {
                        case trouble_1.TROUBLE_METHOD_GUIDE: {
                            try {
                                errTour.go();
                            }
                            catch (e) {
                                throw new ioException_1.IOException(e);
                            }
                            break;
                        }
                        case trouble_1.TROUBLE_METHOD_TEXT: {
                            this.protocolHandler.sendResHeaders(errTour);
                            let data = buffer_1.Buffer.from(cmd.target);
                            errTour.res.sendContent(tour_1.Tour.TOUR_ID_NOCHECK, data, 0, data.length);
                            errTour.res.endContent(tour_1.Tour.TOUR_ID_NOCHECK);
                            break;
                        }
                        case trouble_1.TROUBLE_METHOD_REROUTE: {
                            errTour.res.sendHttpException(tour_1.Tour.TOUR_ID_NOCHECK, httpException_1.HttpException.movedTemp(cmd.target));
                            break;
                        }
                    }
                    handled = true;
                }
            }
        }
        if (!handled) {
            for (const nv of this.portDocker.getAdditionalHeaders()) {
                tur.res.headers.add(nv[0], nv[1]);
            }
            try {
                this.protocolHandler.sendResHeaders(tur);
            }
            catch (e) {
                bayLog_1.BayLog.debug("%s Error on sending headers: %s", tur, e.message);
                tur.changeState(tour_1.Tour.TOUR_ID_NOCHECK, tour_1.Tour.STATE_ABORTED);
                throw e;
            }
        }
    }
    sendRedirect(chkId, tur, status, location) {
        this.checkShipId(chkId);
        let hdr = tur.res.headers;
        hdr.status = status;
        hdr.set(httpHeaders_1.HttpHeaders.LOCATION, location);
        let body = "<H2>Document Moved.</H2><BR>" + "<A HREF=\""
            + location + "\">" + location + "</A>";
        this.sendErrorContent(chkId, tur, body);
    }
    sendResContent(chkId, tur, buf, ofs, len, lis) {
        this.checkShipId(chkId);
        if (tur.isZombie() || tur.isAborted()) {
            // Don't send peer any data. Do nothing
            bayLog_1.BayLog.debug("%s Aborted or zombie tour. do nothing: %s state=%s", this, tur, tur.state);
            tur.changeState(chkId, tour_1.Tour.STATE_ENDED);
            if (lis != null)
                lis();
            return;
        }
        let maxLen = this.protocolHandler.maxResPacketDataSize();
        if (len > maxLen) {
            this.sendResContent(tour_1.Tour.TOUR_ID_NOCHECK, tur, buf, ofs, maxLen, null);
            this.sendResContent(tour_1.Tour.TOUR_ID_NOCHECK, tur, buf, ofs + maxLen, len - maxLen, lis);
        }
        else {
            try {
                this.protocolHandler.sendResContent(tur, buf, ofs, len, lis);
            }
            catch (e) {
                tur.changeState(tour_1.Tour.TOUR_ID_NOCHECK, tour_1.Tour.STATE_ABORTED);
                throw e;
            }
        }
    }
    sendEndTour(chkShipId, chkTourId, tur, callback) {
        this.checkShipId(chkShipId);
        bayLog_1.BayLog.debug("%s sendEndTour: %s state=%s", this, tur, tur.state);
        if (tur.isZombie() || tur.isAborted()) {
            // Don't send peer any data. Only return tour
            bayLog_1.BayLog.debug("%s Aborted or zombie tour. do nothing: %s state=%s", this, tur, tur.state);
            tur.changeState(chkTourId, tour_1.Tour.STATE_ENDED);
            callback();
        }
        else {
            if (!tur.isValid()) {
                throw new sink_1.Sink("Tour is not valid");
            }
            let keepAlive = false;
            if (tur.req.headers.getConnection() == httpHeaders_1.HttpHeaders.CONNECTION_KEEP_ALIVE)
                keepAlive = true;
            if (keepAlive) {
                let resConn = tur.res.headers.getConnection();
                keepAlive = (resConn == httpHeaders_1.HttpHeaders.CONNECTION_KEEP_ALIVE)
                    || (resConn == httpHeaders_1.HttpHeaders.CONNECTION_UNKOWN);
                if (keepAlive) {
                    if (tur.res.headers.contentLength() < 0)
                        keepAlive = false;
                }
            }
            //BayLog.trace("%s sendEndTour: set running false: %s id=%d", this, tur, chkTourId);
            tur.changeState(chkTourId, tour_1.Tour.STATE_ENDED);
            this.protocolHandler.sendEndTour(tur, keepAlive, callback);
        }
    }
    sendError(chkId, tur, status, message, e) {
        this.checkShipId(chkId);
        if (tur == null)
            throw new Error("nullPo");
        bayLog_1.BayLog.debug_e(e, "%s send error: status=%d, message=%s ex=%s", this, status, message, e == null ? "" : e.message);
        if (e != null)
            bayLog_1.BayLog.debug_e(e);
        // Create body
        let str = httpStatus_1.HttpStatus.description(status);
        // print status
        let body = "<h1>" + status.toString() + " " + str + "</h1>\r\n";
        tur.res.headers.status = status;
        this.sendErrorContent(chkId, tur, body);
    }
    //////////////////////////////////////////////////////
    // Custom methods
    //////////////////////////////////////////////////////
    sendErrorContent(chkId, tur, content) {
        // Get charset
        let charset = tur.res.charset;
        // Set content type
        if (charset != null && charset != "") {
            tur.res.headers.setContentType("text/html; charset=" + charset);
        }
        else {
            tur.res.headers.setContentType("text/html");
        }
        let bytes = null;
        if (content != null && content != "") {
            // Create writer
            if (charset != null && charset != "") {
                let encoding = strUtil_1.StrUtil.toEncoding(charset);
                bytes = buffer_1.Buffer.from(content, encoding);
            }
            else {
                bytes = buffer_1.Buffer.from(content);
            }
            tur.res.headers.setContentLength(bytes.length);
        }
        this.sendHeaders(chkId, tur);
        if (bytes != null)
            this.sendResContent(chkId, tur, bytes, 0, bytes.length, null);
        //ship.tourEnded();
    }
}
InboundShip.errCounter = new counter_1.Counter(1);
exports.InboundShip = InboundShip;
//# sourceMappingURL=inboundShip.js.map