"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PortBase = void 0;
const dockerBase_1 = require("./dockerBase");
const inboundShipStore_1 = require("./inboundShipStore");
const ioUtil_1 = require("../../util/ioUtil");
const net = require("net");
const plainTransporter_1 = require("../../agent/transporter/plainTransporter");
const protocolHandlerStore_1 = require("../../protocol/protocolHandlerStore");
const inboundDataListener_1 = require("./inboundDataListener");
const cities_1 = require("../../util/cities");
const strUtil_1 = require("../../util/strUtil");
const configException_1 = require("../../configException");
const bayMessage_1 = require("../../bayMessage");
const symbol_1 = require("../../symbol");
const sysUtil_1 = require("../../util/sysUtil");
const bayLog_1 = require("../../bayLog");
class PortBase extends dockerBase_1.DockerBase {
    constructor() {
        super(...arguments);
        this.permissionList = [];
        this.host = null;
        this.timeoutSec = -1; // -1 means "Use socketTimeout of Harbor docker"
        this.anchored = true;
        this.additionalHeaders = [];
        this.cities = new cities_1.Cities();
    }
    getType() {
        return "port";
    }
    //////////////////////////////////////////////////////
    // Implements Docker
    //////////////////////////////////////////////////////
    init(elm, parent) {
        if (strUtil_1.StrUtil.empty(elm.arg))
            throw new configException_1.ConfigException(elm.fileName, elm.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_INVALID_PORT_NAME, elm.name));
        super.init(elm, parent);
        let portName = elm.arg.toLowerCase();
        if (portName.startsWith(":unix:")) {
            // unix domain sokcet
            if (!sysUtil_1.SysUtil.supportUnixDomainSocketAddress()) {
                throw new configException_1.ConfigException(elm.fileName, elm.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_CANNOT_SUPPORT_UNIX_DOMAIN_SOCKET));
            }
            this.port = -1;
            this.socketPath = elm.arg.substring(6);
            this.host = elm.arg;
        }
        else {
            // TCP or UDP port
            let hostPort;
            if (portName.startsWith(":tcp:")) {
                // tcp server socket
                this.anchored = true;
                hostPort = elm.arg.substring(5);
            }
            else if (portName.startsWith(":udp:")) {
                // udp server socket
                this.anchored = false;
                hostPort = elm.arg.substring(5);
            }
            else {
                // default: tcp server socket
                this.anchored = true;
                hostPort = elm.arg;
            }
            let idx = hostPort.indexOf(':');
            try {
                if (idx >= 0) {
                    this.host = hostPort.substring(0, idx);
                    this.port = Number.parseInt(hostPort.substring(idx + 1));
                }
                else {
                    this.host = null;
                    this.port = Number.parseInt(hostPort);
                }
            }
            catch (e) {
                throw new configException_1.ConfigException(elm.fileName, elm.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_INVALID_PORT_NAME, elm.arg));
            }
        }
        // TCP/UDP support check
        if (this.anchored) {
            if (!this.supportAnchored())
                throw new configException_1.ConfigException(elm.fileName, elm.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_TCP_NOT_SUPPORTED));
        }
        else {
            if (!this.supportUnanchored())
                throw new configException_1.ConfigException(elm.fileName, elm.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_UDP_NOT_SUPPORTED));
        }
    }
    //////////////////////////////////////////////////////
    // Implements DockerBase
    //////////////////////////////////////////////////////
    initDocker(dkr) {
        switch (dkr.getType()) {
            case "permission":
                this.permissionList.push(dkr);
                break;
            case "city":
                this.cities.add(dkr);
                break;
            case "secure":
                this.secureDocker = dkr;
                break;
            default:
                return false;
        }
        return true;
    }
    initKeyVal(kv) {
        switch (kv.key.toLowerCase()) {
            default:
                return super.initKeyVal(kv);
            case "timeout":
                this.timeoutSec = Number.parseInt(kv.value);
                break;
            case "addheader": {
                let idx = kv.value.indexOf(':');
                if (idx < 0) {
                    throw new configException_1.ConfigException(kv.fileName, kv.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_INVALID_PARAMETER_VALUE, kv.value));
                }
                let name = kv.value.substring(0, idx).trim();
                let value = kv.value.substring(idx + 1).trim();
                this.additionalHeaders.push([name, value]);
                break;
            }
        }
        return true;
    }
    //////////////////////////////////////////////////////
    // Implements Port
    //////////////////////////////////////////////////////
    getHost() {
        return this.host;
    }
    getPort() {
        return this.port;
    }
    getSocketPath() {
        return this.socketPath;
    }
    getAddress() {
        return "";
    }
    isAnchored() {
        return this.anchored;
    }
    getTimeoutSec() {
        return this.timeoutSec;
    }
    isSecure() {
        return this.secureDocker != null;
    }
    getAdditionalHeaders() {
        return this.additionalHeaders;
    }
    checkAdmitted(ch) {
        for (const p of this.permissionList)
            p.socketAdmitted(ch);
    }
    getCities() {
        return this.cities.getCities();
    }
    findCity(name) {
        return undefined;
    }
    newTransporter(agent, ch) {
        let sip = PortBase.getShipStore(agent).rent();
        var tp;
        if (this.isSecure())
            tp = this.secureDocker.createTransporter();
        else
            tp = new plainTransporter_1.PlainTransporter(true, ioUtil_1.IOUtil.getSockRecvBufSize(ch.socket));
        let protoHnd = PortBase.getProtocolHandlerStore(this.protocol(), agent).rent();
        sip.initInbound(ch, agent, tp, this, protoHnd);
        tp.init(agent.nonBlockingHandler, ch, new inboundDataListener_1.InboundDataListener(sip));
        return tp;
    }
    returnProtocolHandler(agt, protoHnd) {
        bayLog_1.BayLog.debug("%s Return protocol handler: ", protoHnd);
        PortBase.getProtocolHandlerStore(protoHnd.protocol(), agt).Return(protoHnd);
    }
    returnShip(sip) {
        bayLog_1.BayLog.debug("%s Return ship: ", sip);
        PortBase.getShipStore(sip.agent).Return(sip);
    }
    createServer() {
        if (this.isSecure())
            return this.secureDocker.createServer();
        else
            return new net.Server();
    }
    getSecure() {
        return this.secureDocker;
    }
    //////////////////////////////////////////////////////
    // class methods
    //////////////////////////////////////////////////////
    static getShipStore(agt) {
        return inboundShipStore_1.InboundShipStore.getStore(agt.agentId);
    }
    static getProtocolHandlerStore(protocol, agt) {
        return protocolHandlerStore_1.ProtocolHandlerStore.getStore(protocol, true, agt.agentId);
    }
}
exports.PortBase = PortBase;
//# sourceMappingURL=portBase.js.map