"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BuiltInSecureDocker = void 0;
const dockerBase_1 = require("../base/dockerBase");
const configException_1 = require("../../configException");
const bayMessage_1 = require("../../bayMessage");
const symbol_1 = require("../../symbol");
const strUtil_1 = require("../../util/strUtil");
const bayLog_1 = require("../../bayLog");
const fs = require("fs");
const sysUtil_1 = require("../../util/sysUtil");
const bayserver_1 = require("../../bayserver");
const ioException_1 = require("../../util/ioException");
const tls = require("tls");
const secureTransporter_1 = require("../../agent/transporter/secureTransporter");
class BuiltInSecureDocker extends dockerBase_1.DockerBase {
    constructor() {
        super(...arguments);
        this.clientAuth = BuiltInSecureDocker.DEFAULT_CLIENT_AUTH;
        this.sslProtocol = BuiltInSecureDocker.DEFAULT_SSL_PROTOCOL;
        this.appProtocols = null;
    }
    //////////////////////////////////////////////////////
    // Implements Docker
    //////////////////////////////////////////////////////
    init(elm, parent) {
        super.init(elm, parent);
        if (this.keyStore == null && (this.keyFile == null || this.certFile == null)) {
            throw new configException_1.ConfigException(elm.fileName, elm.lineNo, "Key file or cert file is not specified");
        }
        try {
            this.initSSL();
        }
        catch (e) {
            throw new configException_1.ConfigException(elm.fileName, elm.lineNo, e, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_SSL_INIT_ERROR, e.message), e);
        }
    }
    //////////////////////////////////////////////////////
    // Implements DockerBase
    //////////////////////////////////////////////////////
    initKeyVal(kv) {
        try {
            switch (kv.key.toLowerCase()) {
                default:
                    return false;
                case "key":
                    this.keyFile = this.getFilePath(kv.value);
                    break;
                case "cert":
                    this.certFile = this.getFilePath(kv.value);
                    break;
                case "keystore":
                    this.keyStore = this.getFilePath(kv.value);
                    break;
                case "keystorepass":
                    this.keyStorePass = kv.value;
                    break;
                case "clientauth":
                    this.clientAuth = strUtil_1.StrUtil.parseBool(kv.value);
                    break;
                case "sslprotocol":
                    this.sslProtocol = kv.value;
                    break;
                case "trustcerts":
                    this.certs = this.getFilePath(kv.value);
                    break;
                case "certsPass":
                    this.certsPass = kv.value;
                    break;
                case "tracessl":
                    this.traceSSL = strUtil_1.StrUtil.parseBool(kv.value);
                    break;
            }
            return true;
        }
        catch (e) {
            bayLog_1.BayLog.error_e(e);
            throw new configException_1.ConfigException(kv.fileName, kv.lineNo, bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_FILE_NOT_FOUND, e.message));
        }
    }
    //////////////////////////////////////////////////////
    // Implements Secure
    //////////////////////////////////////////////////////
    setAppProtocols(protocols) {
        this.appProtocols = protocols;
        this.sslctx["ALPNProtocols"] = this.appProtocols;
    }
    createTransporter() {
        return new secureTransporter_1.SecureTransporter(this.sslctx, true, this.traceSSL);
    }
    reloadCert() {
        this.initSSL();
    }
    createServer() {
        return tls.createServer(this.sslctx);
    }
    //////////////////////////////////////////////////////
    // Custom methods
    //////////////////////////////////////////////////////
    initSSL() {
        if (this.traceSSL)
            bayLog_1.BayLog.info("init SSL engine");
        this.sslctx = {
            key: fs.readFileSync(this.keyFile),
            cert: fs.readFileSync(this.certFile),
            requestCert: this.clientAuth,
            enableTrace: this.traceSSL
        };
        if (this.appProtocols)
            this.sslctx["ALPNProtocols"] = this.appProtocols;
    }
    //////////////////////////////////////////////////////
    // Private methods
    //////////////////////////////////////////////////////
    getFilePath(fileName) {
        fileName = bayserver_1.BayServer.getLocation(fileName);
        if (!sysUtil_1.SysUtil.isFile(fileName)) {
            throw new ioException_1.IOException("File not found: " + fileName);
        }
        else
            return fileName;
        return "";
    }
    createTlsSocket(ch, server) {
        let ctx = {
            enableTrace: this.traceSSL,
            isServer: true,
            server: server,
            secureContext: tls.createSecureContext(this.sslctx)
        };
        if (this.appProtocols)
            ctx["ALPNProtocols"] = this.appProtocols;
        return new tls.TLSSocket(ch, ctx);
    }
}
BuiltInSecureDocker.DEFAULT_CLIENT_AUTH = false;
BuiltInSecureDocker.DEFAULT_SSL_PROTOCOL = "TLS";
exports.BuiltInSecureDocker = BuiltInSecureDocker;
module.exports = {
    createDocker: () => new BuiltInSecureDocker()
};
//# sourceMappingURL=builtInSecureDocker.js.map