"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.IpMatcher = void 0;
const net = require("net");
const bayMessage_1 = require("../bayMessage");
const symbol_1 = require("../symbol");
const bayException_1 = require("../bayException");
class IpMatcher {
    constructor(ipDesc) {
        if (ipDesc == "*")
            this.matchAll = true;
        else {
            this.matchAll = false;
            this.parseIp(ipDesc);
        }
    }
    match(adr) {
        if (this.matchAll)
            return true;
        let remoteAdr = IpMatcher.getIpAddress(adr);
        if (remoteAdr.length != this.maskAdr.length)
            return false; // IPv4 and IPv6 don't match each other
        for (let i = 0; i < remoteAdr.length; i++) {
            if ((remoteAdr[i] & this.maskAdr[i]) != this.netAdr[i])
                return false;
        }
        return true;
    }
    parseIp(ipDesc) {
        let ipList = ipDesc.split("/");
        let ip, mask;
        if (ipList.length == 0)
            throw new bayException_1.BayException(bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_INVALID_IP_DESC, ipDesc));
        ip = ipList[0];
        if (ipList.length == 1)
            mask = "255.255.255.255";
        else
            mask = ipList[1];
        let ipAdr = IpMatcher.getIpAddress(ip);
        this.maskAdr = IpMatcher.getIpAddress(mask);
        if (ipAdr.length != this.maskAdr.length) {
            throw new bayException_1.BayException(bayMessage_1.BayMessage.get(symbol_1.Symbol.CFG_IPV4_AND_IPV6_ARE_MIXED, ipDesc));
        }
        this.netAdr = [];
        for (let i = 0; i < this.maskAdr.length; i++)
            this.netAdr[i] = (ipAdr[i] & this.maskAdr[i]);
    }
    static getIpAddress(adr) {
        let ipv6 = false;
        if (adr.indexOf(":") >= 0) {
            if (adr.indexOf(".") > 0) {
                if (!adr.startsWith("::ffff:"))
                    throw new bayException_1.BayException("Invalid IP address: " + adr);
                adr = adr.substring(7);
            }
            else {
                ipv6 = true;
            }
        }
        let sadr = new net.SocketAddress({ address: adr, family: ipv6 ? "ipv6" : "ipv4" });
        if (ipv6) {
            let parts = adr.split("::");
            if (parts.length > 2 || parts.length == 0)
                throw new bayException_1.BayException("Invalid IPv6 address: " + adr);
            if (parts.length == 1)
                return this.parseIPv6Part(parts[0]);
            else {
                let frontPart = this.parseIPv6Part(parts[0]);
                let backPart = this.parseIPv6Part(parts[1]);
                while (frontPart.length + backPart.length < 16)
                    frontPart.push(0);
                return [...frontPart, ...backPart];
            }
        }
        else {
            return adr.split(".").map((item) => {
                return Number.parseInt(item);
            });
        }
    }
    static parseIPv6Part(part) {
        let items = part.split(":");
        let address = [];
        for (const item of items) {
            if (item == "")
                continue;
            let val = Number.parseInt(item, 16);
            address.push(val >> 8);
            address.push(val & 0xff);
        }
        return address;
    }
}
exports.IpMatcher = IpMatcher;
//# sourceMappingURL=ipMatcher.js.map